import copy
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Dict,
    Generator,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

import sqlalchemy

from edgy.conf import settings
from edgy.core.db.context_vars import get_schema
from edgy.core.db.fields import CharField, TextField
from edgy.core.db.fields.base import BaseForeignKey, RelationshipField
from edgy.core.db.querysets.mixins import EdgyModel, QuerySetPropsMixin, TenancyMixin
from edgy.core.db.querysets.prefetch import PrefetchMixin
from edgy.core.db.querysets.protocols import AwaitableQuery
from edgy.core.utils.models import DateParser, ModelParser
from edgy.exceptions import MultipleObjectsReturned, ObjectNotFound, QuerySetError
from edgy.protocols.queryset import QuerySetProtocol

from . import clauses as clauses_mod

if TYPE_CHECKING:  # pragma: no cover
    from edgy import Database
    from edgy.core.db.models import Model


class RelationshipCrawlResult(NamedTuple):
    model_class: Type["Model"]
    field_name: str
    operator: str
    forward_path: str
    reverse_path: Union[str, Literal[False]]

def crawl_relationship(model_class: Type["Model"], path: str, callback_fn: Any=None) -> RelationshipCrawlResult:
    field = None
    forward_prefix_path = ""
    reverse_path: Union[str, Literal[False]] = ""
    operator: str = "exact"
    field_name: str = path
    while path:
        splitted = path.split("__", 1)
        field_name = splitted[0]
        field = model_class.meta.fields_mapping.get(field_name)
        if isinstance(field, RelationshipField) and len(splitted) == 2:
            model_class, reverse_part, path = field.traverse_field(path)
            if field.is_cross_db():
                raise NotImplementedError("We cannot cross databases yet, this feature is planned")
            reverse = not isinstance(field, BaseForeignKey)
            if reverse_part and reverse_path is not False:
                if reverse_path:
                    reverse_path = f"{reverse_part}__{reverse_path}"
                else:
                    reverse_path = reverse_part
            else:
                reverse_path = False

            if callback_fn:
                callback_fn(model_class=model_class, field=field, reverse_path=reverse_path, forward_path=forward_prefix_path, reverse=reverse, operator=None)
            if forward_prefix_path:
                forward_prefix_path =  f"{forward_prefix_path}__{field_name}"
            else:
                forward_prefix_path = field_name
        elif len(splitted) == 2:
            if "__" not in splitted[1] and splitted[1] in settings.filter_operators:
                operator = splitted[1]
                break
            else:
                raise ValueError(f"Tried to cross field: {field_name} of type {field!r}, remainder: {splitted[1]}")
        else:
            operator = "exact"
            break

    if reverse_path is not False:
        if reverse_path:
            reverse_path = f"{field_name}__{reverse_path}"
        else:
            reverse_path = field_name
    if callback_fn and field is not None:
        callback_fn(model_class=model_class, field=field, reverse_path=reverse_path, forward_path=forward_prefix_path, reverse=False, operator=operator)
    return RelationshipCrawlResult(
        model_class=model_class,
        field_name=field_name,
        operator=operator,
        forward_path=forward_prefix_path,
        reverse_path=reverse_path,
    )

def clean_query_kwargs(model: Type["Model"], kwargs: Dict[str, Any]) -> Dict[str, Any]:
    new_kwargs: Dict[str, Any] = {}
    for key, val in kwargs.items():
        model_class, field_name, _, _, _ = crawl_relationship(model, key)
        field = model_class.meta.fields_mapping.get(field_name)
        if field is not None:
            new_kwargs.update(field.clean(key, val, for_query=True))
        else:
            new_kwargs[key] = val
    assert "pk" not in new_kwargs, "pk should be already parsed"
    return new_kwargs


class BaseQuerySet(
    TenancyMixin,
    QuerySetPropsMixin,
    PrefetchMixin,
    DateParser,
    ModelParser,
    AwaitableQuery[EdgyModel],
):
    ESCAPE_CHARACTERS = ["%", "_"]

    def __init__(
        self,
        model_class: Union[Type["Model"], None] = None,
        database: Union["Database", None] = None,
        filter_clauses: Any = None,
        or_clauses: Any = None,
        select_related: Any = None,
        prefetch_related: Any = None,
        limit_count: Any = None,
        limit_offset: Any = None,
        order_by: Any = None,
        group_by: Any = None,
        distinct_on: Optional[Sequence[str]] = None,
        only_fields: Any = None,
        defer_fields: Any = None,
        embed_parent: Any = None,
        using_schema: Any = None,
        table: Any = None,
        exclude_secrets: Any = False,
    ) -> None:
        super().__init__(model_class=model_class)
        self.model_class = cast("Type[Model]", model_class)
        self.filter_clauses = [] if filter_clauses is None else filter_clauses
        self.or_clauses = [] if or_clauses is None else or_clauses
        self.limit_count = limit_count
        self._select_related = [] if select_related is None else select_related
        self._prefetch_related = [] if prefetch_related is None else prefetch_related
        self._offset = limit_offset
        self._order_by = [] if order_by is None else order_by
        self._group_by = [] if group_by is None else group_by
        self.distinct_on = distinct_on
        self._only = [] if only_fields is None else only_fields
        self._defer = [] if defer_fields is None else defer_fields
        self._expression = None
        self._cache = None
        self.embed_parent = embed_parent
        self.using_schema = using_schema
        self._exclude_secrets = exclude_secrets or False
        self.extra: Dict[str, Any] = {}

        # Making sure the queryset always starts without any schema associated unless specified

        if table is not None:
            self.table = table
        if database is not None:
            self.database = database

    def _build_order_by_expression(self, order_by: Any, expression: Any) -> Any:
        """Builds the order by expression"""
        order_by = list(map(self._prepare_order_by, order_by))
        expression = expression.order_by(*order_by)
        return expression

    def _build_group_by_expression(self, group_by: Any, expression: Any) -> Any:
        """Builds the group by expression"""
        group_by = list(map(self._prepare_group_by, group_by))
        expression = expression.group_by(*group_by)
        return expression

    def _build_filter_clauses_expression(self, filter_clauses: Any, expression: Any) -> Any:
        """Builds the filter clauses expression"""
        if len(filter_clauses) == 1:
            clause = filter_clauses[0]
        else:
            clause = clauses_mod.and_(*filter_clauses)
        expression = expression.where(clause)
        return expression

    def _build_or_clauses_expression(self, or_clauses: Any, expression: Any) -> Any:
        """Builds the filter clauses expression"""
        if len(or_clauses) == 1:
            clause = or_clauses[0]
        else:
            clause = clauses_mod.or_(*or_clauses)
        expression = expression.where(clause)
        return expression

    def _build_select_distinct(self, distinct_on: Optional[Sequence[str]], expression: Any) -> Any:
        """Filters selects only specific fields. Leave empty to use simple distinct"""
        # using with columns is not supported by all databases
        if distinct_on:
            return expression.distinct(*map(self._prepare_fields_for_distinct, distinct_on))
        else:
            return expression.distinct()

    def _build_tables_select_from_relationship(self) -> Any:
        """
        Builds the tables relationships and joins.
        When a table contains more than one foreign key pointing to the same
        destination table, a lookup for the related field is made to understand
        from which foreign key the table is looked up from.
        """
        queryset: "QuerySet" = self._clone()

        select_from = queryset.table
        tables = {select_from.name: select_from}

        # Select related
        for select_path in queryset._select_related:
            # For m2m relationships
            model_class = queryset.model_class
            select_from = queryset.table
            former_table = None
            while select_path:
                field_name = select_path.split("__", 1)[0]
                field = model_class.fields[field_name]
                if isinstance(field, RelationshipField):
                    model_class, reverse_part, select_path = field.traverse_field(select_path)
                else:
                    raise ValueError(f"{field_name}: invalid field type: {field!r}")
                if isinstance(field, BaseForeignKey):
                    foreign_key = field
                    reverse = False
                else:
                    foreign_key = model_class.meta.fields_mapping[reverse_part]
                    reverse = True
                if foreign_key.is_cross_db():
                    raise NotImplementedError("We cannot cross databases yet, this feature is planned")
                table = model_class.table
                select_from = sqlalchemy.sql.join(  # type: ignore
                    select_from,
                    table,
                    *self._select_from_relationship_clause_generator(select_from, foreign_key, table, reverse, former_table)
                )
                former_table = table
                tables[table.name] = table

        return tables.values(), select_from

    @staticmethod
    def _select_from_relationship_clause_generator(select_from: Any, foreign_key: BaseForeignKey, table: Any, reverse: bool, former_table: Any=None) -> Any:
        column_names = foreign_key.get_column_names(foreign_key.name)
        for col in column_names:
            if reverse:
                colname = foreign_key.from_fk_field_name(foreign_key.name, col)
            else:
                colname = col
            if former_table is None:
                former_table = select_from
            if reverse:
                yield getattr(former_table.c, colname) == getattr(table.c, col)
            else:
                yield getattr(former_table.c, colname) == getattr(table.c, foreign_key.from_fk_field_name(foreign_key.name, col))



    def _validate_only_and_defer(self) -> None:
        if self._only and self._defer:
            raise QuerySetError("You cannot use .only() and .defer() at the same time.")

    def _secret_recursive_names(self, model_class: Any, columns: Union[List[str], None] = None) -> List[str]:
        """
        Recursively gets the names of the fields excluding the secrets.
        """
        if columns is None:
            columns = []

        for name, field in model_class.fields.items():
            if isinstance(field, BaseForeignKey):
                # Making sure the foreign key is always added unless is a secret
                if not field.secret:
                    columns.extend(field.get_column_names(name))
                    columns.extend(self._secret_recursive_names(model_class=field.target, columns=columns))
                continue
            if not field.secret:
                columns.append(name)

        columns = list(set(columns))
        return columns

    def _build_select(self) -> Any:
        """
        Builds the query select based on the given parameters and filters.
        """
        queryset: "QuerySet" = self._clone()

        queryset._validate_only_and_defer()
        tables, select_from = queryset._build_tables_select_from_relationship()
        expression = sqlalchemy.sql.select(*tables)
        expression = expression.select_from(select_from)

        if queryset._only:
            expression = expression.with_only_columns(*queryset._only)

        if queryset._defer:
            columns = [column for column in select_from.columns if column.name not in queryset._defer]
            expression = expression.with_only_columns(*columns)

        if queryset._exclude_secrets:
            model_columns = queryset._secret_recursive_names(model_class=queryset.model_class)
            columns = [column for column in select_from.columns if column.name in model_columns]
            expression = expression.with_only_columns(*columns)

        if queryset.filter_clauses:
            expression = queryset._build_filter_clauses_expression(queryset.filter_clauses, expression=expression)

        if queryset.or_clauses:
            expression = queryset._build_or_clauses_expression(queryset.or_clauses, expression=expression)

        if queryset._order_by:
            expression = queryset._build_order_by_expression(queryset._order_by, expression=expression)

        if queryset.limit_count:
            expression = expression.limit(queryset.limit_count)

        if queryset._offset:
            expression = expression.offset(queryset._offset)

        if queryset._group_by:
            expression = queryset._build_group_by_expression(queryset._group_by, expression=expression)

        if queryset.distinct_on is not None:
            expression = queryset._build_select_distinct(queryset.distinct_on, expression=expression)

        queryset._expression = expression  # type: ignore
        return expression

    def _filter_query(
        self,
        exclude: bool = False,
        or_: bool = False,
        **kwargs: Any,
    ) -> "QuerySet":
        from edgy.core.db.models import Model

        clauses = []
        filter_clauses = self.filter_clauses
        or_clauses = self.or_clauses
        select_related = list(self._select_related)
        prefetch_related = list(self._prefetch_related)

        # Making sure for queries we use the main class and not the proxy
        # And enable the parent
        if self.model_class.is_proxy_model:
            self.model_class = self.model_class.parent

        kwargs = clean_query_kwargs(self.model_class, kwargs)

        for key, value in kwargs.items():
            assert not isinstance(value, Model), f"should be parsed in clean: {key}: {value}"
            model_class, field_name, op, related_str, _ = crawl_relationship(self.model_class, key)
            if related_str and related_str not in select_related:
                select_related.append(related_str)
            column = model_class.table.columns[field_name]

            # Map the operation code onto SQLAlchemy's ColumnElement
            # https://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.ColumnElement
            op_attr = settings.filter_operators[op]
            has_escaped_character = False

            if op in ["contains", "icontains"]:
                has_escaped_character = any(c for c in self.ESCAPE_CHARACTERS if c in value)
                if has_escaped_character:
                    # enable escape modifier
                    for char in self.ESCAPE_CHARACTERS:
                        value = value.replace(char, f"\\{char}")
                value = f"%{value}%"

            clause = getattr(column, op_attr)(value)
            clause.modifiers["escape"] = "\\" if has_escaped_character else None
            clauses.append(clause)

        if exclude:
            if not or_:
                filter_clauses.append(clauses_mod.not_(clauses_mod.and_(*clauses)))
            else:
                or_clauses.append(clauses_mod.not_(clauses_mod.and_(*clauses)))
        else:
            if not or_:
                filter_clauses += clauses
            else:
                or_clauses += clauses

        return cast(
            "QuerySet",
            self.__class__(
                model_class=self.model_class,
                database=self._database,
                filter_clauses=filter_clauses,
                or_clauses=or_clauses,
                select_related=select_related,
                prefetch_related=prefetch_related,
                limit_count=self.limit_count,
                limit_offset=self._offset,
                order_by=self._order_by,
                only_fields=self._only,
                defer_fields=self._defer,
                embed_parent=self.embed_parent,
                table=self.table,
                exclude_secrets=self._exclude_secrets,
                using_schema=self.using_schema,
            ),
        )

    def _validate_kwargs(self, **kwargs: Any) -> Any:
        return self._extract_values_from_field(kwargs, model_class=self.model_class)

    def _prepare_order_by(self, order_by: str) -> Any:
        reverse = order_by.startswith("-")
        order_by = order_by.lstrip("-")
        order_col = self.table.columns[order_by]
        return order_col.desc() if reverse else order_col

    def _prepare_group_by(self, group_by: str) -> Any:
        group_by = group_by.lstrip("-")
        group_col = self.table.columns[group_by]
        return group_col

    def _prepare_fields_for_distinct(self, distinct_on: str) -> sqlalchemy.Column:
        return self.table.columns[distinct_on]

    def _clone(self) -> Any:
        """
        Return a copy of the current QuerySet that's ready for another
        operation.
        """
        queryset = self.__class__.__new__(self.__class__)
        queryset.model_class = self.model_class

        # Making sure the registry schema takes precendent with
        # Any provided using
        if not self.model_class.meta.registry.db_schema:
            schema = get_schema()
            if self.using_schema is None and schema is not None:
                self.using_schema = schema
            queryset.model_class.table = self.model_class.build(self.using_schema)

        queryset.filter_clauses = copy.copy(self.filter_clauses)
        queryset.or_clauses = copy.copy(self.or_clauses)
        queryset.limit_count = copy.copy(self.limit_count)
        queryset._select_related = copy.copy(self._select_related)
        queryset._prefetch_related = copy.copy(self._prefetch_related)
        queryset._offset = copy.copy(self._offset)
        queryset._order_by = copy.copy(self._order_by)
        queryset._group_by = copy.copy(self._group_by)
        queryset.distinct_on = copy.copy(self.distinct_on)
        queryset._expression = copy.copy(self._expression)
        queryset.embed_parent = self.embed_parent
        queryset._only = copy.copy(self._only)
        queryset._defer = copy.copy(self._defer)
        queryset._database = self.database
        queryset.table = self.table
        queryset.extra = self.extra
        queryset._exclude_secrets = self._exclude_secrets
        queryset.using_schema = self.using_schema

        return queryset


class QuerySet(BaseQuerySet, QuerySetProtocol):
    """
    QuerySet object used for query retrieving.
    """

    def __get__(self, instance: Any, owner: Any = None) -> "QuerySet":
        return self.__class__(model_class=owner if owner else instance.__class__)

    @property
    def sql(self) -> str:
        return str(self._expression)

    @sql.setter
    def sql(self, value: Any) -> None:
        self._expression = value

    async def __aiter__(self) -> AsyncIterator[EdgyModel]:
        for value in await self:
            yield value

    def _set_query_expression(self, expression: Any) -> None:
        """
        Sets the value of the sql property to the expression used.
        """
        self.sql = expression
        self.model_class.raw_query = self.sql

    def _filter_or_exclude(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        exclude: bool = False,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Filters or excludes a given clause for a specific QuerySet.
        """
        queryset: "QuerySet" = self._clone()
        if clause is None:
            return queryset._filter_query(exclude=exclude, **kwargs)

        queryset.filter_clauses.append(clause)
        return queryset

    def filter(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Filters the QuerySet by the given kwargs and clause.
        """
        return self._filter_or_exclude(clause=clause, **kwargs)

    def or_(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Filters the QuerySet by the OR operand.
        """
        queryset: "QuerySet" = self._clone()
        queryset = self.filter(clause=clause, or_=True, **kwargs)
        return queryset

    def and_(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Filters the QuerySet by the AND operand.
        """
        queryset: "QuerySet" = self._clone()
        queryset = self.filter(clause=clause, **kwargs)
        return queryset

    def not_(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Filters the QuerySet by the NOT operand.
        """
        queryset: "QuerySet" = self._clone()
        queryset = queryset.exclude(clause=clause, **kwargs)
        return queryset

    def exclude(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Exactly the same as the filter but for the exclude.
        """
        queryset: "QuerySet" = self._clone()
        queryset = self._filter_or_exclude(clause=clause, exclude=True, **kwargs)
        return queryset

    def exclude_secrets(
        self,
        clause: Optional[sqlalchemy.sql.expression.BinaryExpression] = None,
        **kwargs: Any,
    ) -> "QuerySet":
        """
        Excludes any field that contains the `secret=True` declared from being leaked.
        """
        queryset: "QuerySet" = self._clone()
        queryset._exclude_secrets = True
        queryset = queryset.filter(clause=clause, **kwargs)
        return queryset

    def lookup(self, term: Any) -> "QuerySet":
        """
        Broader way of searching for a given term
        """
        queryset: "QuerySet" = self._clone()
        if not term:
            return queryset

        filter_clauses = list(queryset.filter_clauses)
        value = f"%{term}%"

        search_fields = [
            name for name, field in queryset.model_class.fields.items() if isinstance(field, (CharField, TextField))
        ]
        search_clauses = [queryset.table.columns[name].ilike(value) for name in search_fields]

        if len(search_clauses) > 1:
            filter_clauses.append(sqlalchemy.sql.or_(*search_clauses))
        else:
            filter_clauses.extend(search_clauses)

        return queryset

    def order_by(self, *order_by: str) -> "QuerySet":
        """
        Returns a QuerySet ordered by the given fields.
        """
        queryset: "QuerySet" = self._clone()
        queryset._order_by = order_by
        return queryset

    def limit(self, limit_count: int) -> "QuerySet":
        """
        Returns a QuerySet limited by.
        """
        queryset: "QuerySet" = self._clone()
        queryset.limit_count = limit_count
        return queryset

    def offset(self, offset: int) -> "QuerySet":
        """
        Returns a Queryset limited by the offset.
        """
        queryset: "QuerySet" = self._clone()
        queryset._offset = offset
        return queryset

    def group_by(self, *group_by: Sequence[str]) -> "QuerySet":
        """
        Returns the values grouped by the given fields.
        """
        queryset: "QuerySet" = self._clone()
        queryset._group_by = group_by
        return queryset

    def distinct(self, *distinct_on: str) -> "QuerySet":
        """
        Returns a queryset with distinct results.
        """
        queryset: "QuerySet" = self._clone()
        queryset.distinct_on = distinct_on
        return queryset

    def only(self, *fields: Sequence[str]) -> "QuerySet":
        """
        Returns a list of models with the selected only fields and always the primary
        key.
        """
        only_fields = [sqlalchemy.text(field) for field in fields]
        missing = []
        if self.model_class.pknames:
            for pkname in self.model_class.pknames:
                if pkname not in fields:
                    for pkcolumn in self.model_class.meta.get_columns_for_name(pkname):
                        missing.append(sqlalchemy.text(pkcolumn.key))
        else:
            for pkcolumn in self.model_class.pkcolumns:
                missing.append(sqlalchemy.text(pkcolumn.key))
        if missing:
            only_fields = missing + only_fields

        queryset: "QuerySet" = self._clone()
        queryset._only = only_fields
        return queryset

    def defer(self, *fields: Sequence[str]) -> "QuerySet":
        """
        Returns a list of models with the selected only fields and always the primary
        key.
        """
        queryset: "QuerySet" = self._clone()
        queryset._defer = fields
        return queryset

    def select_related(self, related: Any) -> "QuerySet":
        """
        Returns a QuerySet that will “follow” foreign-key relationships, selecting additional
        related-object data when it executes its query.

        This is a performance booster which results in a single more complex query but means

        later use of foreign-key relationships won’t require database queries.
        """
        queryset: "QuerySet" = self._clone()
        if not isinstance(related, (list, tuple)):
            related = [related]

        related = list(queryset._select_related) + related
        queryset._select_related = related
        return queryset

    async def values(
        self,
        fields: Union[Sequence[str], str, None] = None,
        exclude: Union[Sequence[str], Set[str]] = None,
        exclude_none: bool = False,
        flatten: bool = False,
        **kwargs: Any,
    ) -> List[Any]:
        """
        Returns the results in a python dictionary format.
        """
        fields = fields or []
        queryset: "QuerySet" = self._clone()
        rows: List["Model"] = await queryset.all()

        if not isinstance(fields, list):
            raise QuerySetError(detail="Fields must be an iterable.")

        if not fields:
            rows = [row.model_dump(exclude=exclude, exclude_none=exclude_none) for row in rows]
        else:
            rows = [row.model_dump(exclude=exclude, exclude_none=exclude_none, include=fields) for row in rows]

        as_tuple = kwargs.pop("__as_tuple__", False)

        if not as_tuple:
            return rows

        if not flatten:
            rows = [tuple(row.values()) for row in rows]  # type: ignore
        else:
            try:
                rows = [row[fields[0]] for row in rows]  # type: ignore
            except KeyError:
                raise QuerySetError(detail=f"{fields[0]} does not exist in the results.") from None
        return rows

    async def values_list(
        self,
        fields: Union[Sequence[str], str, None] = None,
        exclude: Union[Sequence[str], Set[str]] = None,
        exclude_none: bool = False,
        flat: bool = False,
    ) -> List[Any]:
        """
        Returns the results in a python dictionary format.
        """
        queryset: "QuerySet" = self._clone()
        fields = fields or []
        if flat and len(fields) > 1:
            raise QuerySetError(
                detail=f"Maximum of 1 in fields when `flat` is enables, got {len(fields)} instead."
            ) from None

        if flat and isinstance(fields, str):
            fields = [fields]

        if isinstance(fields, str):
            fields = [fields]

        return await queryset.values(
            fields=fields,
            exclude=exclude,
            exclude_none=exclude_none,
            flatten=flat,
            __as_tuple__=True,
        )

    async def exists(self, **kwargs: Any) -> bool:
        """
        Returns a boolean indicating if a record exists or not.
        """
        queryset: "QuerySet" = self._clone()
        expression = queryset._build_select()
        expression = sqlalchemy.exists(expression).select()
        queryset._set_query_expression(expression)
        _exists = await queryset.database.fetch_val(expression)
        return cast("bool", _exists)

    async def count(self, **kwargs: Any) -> int:
        """
        Returns an indicating the total records.
        """
        queryset: "QuerySet" = self._clone()
        expression = queryset._build_select().alias("subquery_for_count")
        expression = sqlalchemy.func.count().select().select_from(expression)
        queryset._set_query_expression(expression)
        _count = await queryset.database.fetch_val(expression)
        return cast("int", _count)

    async def get_or_none(self, **kwargs: Any) -> Union[EdgyModel, None]:
        """
        Fetch one object matching the parameters or returns None.
        """
        queryset: "QuerySet" = self.filter(**kwargs)
        expression = queryset._build_select().limit(2)
        queryset._set_query_expression(expression)
        rows = await queryset.database.fetch_all(expression)

        if not rows:
            return None
        if len(rows) > 1:
            raise MultipleObjectsReturned()
        return queryset.model_class.from_sqla_row(
            rows[0],
            select_related=queryset._select_related,
            exclude_secrets=queryset._exclude_secrets,
            using_schema=queryset.using_schema,
        )

    def embed_parent_in_result(self, result: Any) -> Any:
        if not self.embed_parent:
            return result
        new_result = getattr(result, self.embed_parent[0])
        if self.embed_parent[1]:
            setattr(new_result, self.embed_parent[1], result)
        return new_result

    async def _all(self, **kwargs: Any) -> List[EdgyModel]:
        """
        Executes the query.
        """
        queryset: "QuerySet" = self._clone()
        if queryset.embed_parent:
            # activates distinct, not distinct on
            queryset.distinct_on = []

        if kwargs:
            return await queryset.filter(**kwargs).all()

        expression = queryset._build_select()
        queryset._set_query_expression(expression)

        rows = await queryset.database.fetch_all(expression)

        # Attach the raw query to the object
        queryset.model_class.raw_query = queryset.sql

        is_only_fields = True if queryset._only else False
        is_defer_fields = True if queryset._defer else False

        results = [
            queryset.model_class.from_sqla_row(
                row,
                select_related=queryset._select_related,
                prefetch_related=queryset._prefetch_related,
                is_only_fields=is_only_fields,
                only_fields=queryset._only,
                is_defer_fields=is_defer_fields,
                exclude_secrets=queryset._exclude_secrets,
                using_schema=queryset.using_schema,
            )
            for row in rows
        ]

        all_results = [self.embed_parent_in_result(result) for result in results]
        return all_results

    def all(self, **kwargs: Any) -> "QuerySet":
        """
        Returns the queryset records based on specific filters
        """
        queryset: "QuerySet" = self._clone()
        queryset.extra = kwargs
        return queryset

    async def get(self, **kwargs: Any) -> EdgyModel:
        """
        Returns a single record based on the given kwargs.
        """
        queryset: "QuerySet" = self._clone()

        if kwargs:
            return await queryset.filter(**kwargs).get()

        expression = queryset._build_select().limit(2)
        rows = await queryset.database.fetch_all(expression)
        queryset._set_query_expression(expression)

        is_only_fields = True if queryset._only else False
        is_defer_fields = True if queryset._defer else False

        if not rows:
            raise ObjectNotFound()
        if len(rows) > 1:
            raise MultipleObjectsReturned()

        return self.embed_parent_in_result(queryset.model_class.from_sqla_row(
            rows[0],
            select_related=queryset._select_related,
            is_only_fields=is_only_fields,
            only_fields=queryset._only,
            is_defer_fields=is_defer_fields,
            prefetch_related=queryset._prefetch_related,
            exclude_secrets=queryset._exclude_secrets,
            using_schema=queryset.using_schema,
        ))

    async def first(self, **kwargs: Any) -> Union[EdgyModel, None]:
        """
        Returns the first record of a given queryset.
        """
        queryset: "QuerySet" = self._clone()
        if kwargs:
            return await queryset.filter(**kwargs).order_by("id").get()

        rows = await queryset.limit(1).order_by("id").all()
        if rows:
            return self.embed_parent_in_result(rows[0])
        return None

    async def last(self, **kwargs: Any) -> Union[EdgyModel, None]:
        """
        Returns the last record of a given queryset.
        """
        queryset: "QuerySet" = self._clone()
        if kwargs:
            return await queryset.filter(**kwargs).order_by("-id").get()

        rows = await queryset.order_by("-id").all()
        if rows:
            return self.embed_parent_in_result(rows[0])
        return None

    async def create(self, **kwargs: Any) -> EdgyModel:
        """
        Creates a record in a specific table.
        """
        queryset: "QuerySet" = self._clone()
        instance = queryset.model_class(**kwargs)
        instance.table = queryset.table
        instance = await instance.save(force_save=True)
        return self.embed_parent_in_result(instance)

    async def bulk_create(self, objs: List[Dict]) -> None:
        """
        Bulk creates records in a table
        """
        queryset: "QuerySet" = self._clone()
        new_objs = [queryset._validate_kwargs(**obj) for obj in objs]

        expression = queryset.table.insert().values(new_objs)
        queryset._set_query_expression(expression)
        await queryset.database.execute(expression)

    async def bulk_update(self, objs: List[EdgyModel], fields: List[str]) -> None:
        """
        Bulk updates records in a table.

        A similar solution was suggested here: https://github.com/encode/orm/pull/148

        It is thought to be a clean approach to a simple problem so it was added here and
        refactored to be compatible with Edgy.
        """
        queryset: "QuerySet" = self._clone()

        new_objs = []
        for obj in objs:
            new_obj = {}
            for key, value in obj.__dict__.items():
                if key in fields:
                    new_obj[key] = self._resolve_value(value)
            new_objs.append(new_obj)

        new_objs = [queryset._extract_values_from_field(obj, queryset.model_class) for obj in new_objs]

        pks1 = (getattr(queryset.table.c, pkcol) == sqlalchemy.bindparam(pkcol) for pkcol in queryset.pkcolumns)
        expression = queryset.table.update().where(*pks1)
        kwargs: Dict[Any, Any] = {field: sqlalchemy.bindparam(field) for obj in new_objs for field in obj.keys()}
        pks2 = [{pkcol: getattr(obj, pkcol) for pkcol in queryset.pkcolumns} for obj in objs]

        query_list = []
        for pk, value in zip(pks2, new_objs):  # noqa
            query_list.append({**pk, **value})

        expression = expression.values(kwargs)
        queryset._set_query_expression(expression)
        await queryset.database.execute_many(str(expression), query_list)

    async def delete(self) -> None:
        queryset: "QuerySet" = self._clone()

        await self.model_class.signals.pre_delete.send_async(self.__class__, instance=self)

        expression = queryset.table.delete()
        for filter_clause in queryset.filter_clauses:
            expression = expression.where(filter_clause)

        queryset._set_query_expression(expression)
        await queryset.database.execute(expression)

        await self.model_class.signals.post_delete.send_async(self.__class__, instance=self)

    async def update(self, **kwargs: Any) -> None:
        """
        Updates a record in a specific table with the given kwargs.
        """
        queryset: "QuerySet" = self._clone()

        extracted_fields = queryset._extract_values_from_field(kwargs, model_class=queryset.model_class)
        kwargs = queryset._update_auto_now_fields(extracted_fields, queryset.model_class.fields)

        # Broadcast the initial update details
        await self.model_class.signals.pre_update.send_async(self.__class__, instance=self, kwargs=kwargs)

        expression = queryset.table.update().values(**kwargs)

        for filter_clause in queryset.filter_clauses:
            expression = expression.where(filter_clause)

        queryset._set_query_expression(expression)
        await queryset.database.execute(expression)

        # Broadcast the update executed
        await self.model_class.signals.post_update.send_async(self.__class__, instance=self)

    async def get_or_create(self, defaults: Dict[str, Any], **kwargs: Any) -> Tuple[EdgyModel, bool]:
        """
        Creates a record in a specific table or updates if already exists.
        """
        queryset: "QuerySet" = self._clone()

        try:
            instance = await queryset.get(**kwargs)
            return instance, False
        except ObjectNotFound:
            kwargs.update(defaults)
            instance = await queryset.create(**kwargs)
            return instance, True

    async def update_or_create(self, defaults: Dict[str, Any], **kwargs: Any) -> Tuple[EdgyModel, bool]:
        """
        Updates a record in a specific table or creates a new one.
        """
        queryset: "QuerySet" = self._clone()
        try:
            instance = await queryset.get(**kwargs)
            await instance.update(**defaults)
            return instance, False
        except ObjectNotFound:
            kwargs.update(defaults)
            instance = await queryset.create(**kwargs)
            return instance, True

    async def contains(self, instance: EdgyModel) -> bool:
        """Returns true if the QuerySet contains the provided object.
        False if otherwise.
        """
        queryset: "QuerySet" = self._clone()

        if getattr(instance, "pk", None) is None:
            raise ValueError("'obj' must be a model or reflect model instance.")
        # TODO: handle embed parent
        return await queryset.filter(pk=instance.pk).exists()

    async def _execute(self) -> Any:
        queryset: "QuerySet" = self._clone()
        records = await queryset._all(**queryset.extra)
        return records

    def __await__(
        self,
    ) -> Generator[Any, None, List[EdgyModel]]:
        return self._execute().__await__()

    def __class_getitem__(cls, *args: Any, **kwargs: Any) -> Any:
        return cls



import functools
from typing import TYPE_CHECKING, Any, Literal, Optional, Sequence, Tuple, Type, Union, cast

import sqlalchemy
from pydantic import BaseModel, ConfigDict

from edgy.core.db.fields.base import RelationshipField
from edgy.exceptions import ObjectNotFound, RelationshipIncompatible, RelationshipNotFound
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:
    from edgy import Model, QuerySet, ReflectModel



def _removeprefix(text: str, prefix: str) -> str:
    # TODO: replace with removeprefix when python3.9 is minimum
    if text.startswith(prefix):
        return text[len(prefix) :]
    else:
        return text


def _removeprefixes(text:str, *prefixes: Sequence[str]) -> str:
    for prefix in prefixes:
        text = _removeprefix(text, prefix)
    return text


class ManyRelation(ManyRelationProtocol):
    """
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(
        self,
        *,
        from_foreign_key: str,
        to_foreign_key: str,
        to: Union[Type["Model"], Type["ReflectModel"]],
        through: Union[Type["Model"], Type["ReflectModel"]],
        reverse: bool=False,
        embed_through: Union[Literal[False], str] = "",
        refs: Any = (),
        instance: Optional[Union["Model", "ReflectModel"]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.through = through
        self.to = to
        self.instance = instance
        self.reverse = reverse
        self.from_foreign_key = from_foreign_key
        self.to_foreign_key = to_foreign_key
        self.embed_through = embed_through
        self.refs: Sequence[Union["Model", "ReflectModel"]] = []
        if not isinstance(refs, Sequence):
            refs = [refs]
        for v in refs:
            self._add_object(v)

    def get_queryset(self) -> "QuerySet":
        # we need to check tenant every request
        queryset = self.through.meta.managers["query_related"].get_queryset()
        queryset.embed_parent = (self.to_foreign_key, self.embed_through)
        return queryset

    async def save_related(self) -> None:
        # TODO: improve performance
        fk = self.through.meta.fields_mapping[self.from_foreign_key]
        while self.refs:
            ref = self.refs.pop()
            ref.__dict__.update(fk.clean(fk.name, self.instance))
            await self.add(ref)

    def __getattr__(self, item: Any) -> Any:
        """
        Gets the attribute from the queryset and if it does not
        exist, then lookup in the model.
        """#
        try:
            attr = getattr(self.get_queryset(), item)
        except AttributeError:
            attr = getattr(self.through, item)

        func = self.wrap_args(attr)
        return func

    def wrap_args(self, func: Any) -> Any:
        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            assert self.instance, "instance not initialized"
            fk = self.through.meta.fields_mapping[self.from_foreign_key]
            query = {}
            if self.embed_through == "":
                new_kwargs = kwargs
            else:
                new_kwargs = {}
                for key, val in kwargs.items():
                    if self.embed_through is False or not key.startswith(self.embed_through):
                        new_kwargs[f"{self.to_foreign_key}__{key}"] = val
                    else:
                        new_kwargs[_removeprefixes(key, self.embed_through, "__")] = val
            for related_name in fk.related_columns.keys():
                query[related_name] = getattr(self.instance, related_name)
            new_kwargs[self.from_foreign_key] = query
            return func(*args, **new_kwargs)

        return wrapped

    def expand_relationship(self, value: Any) -> Any:
        through = self.through

        if isinstance(value, (through, through.proxy_model)):
            return value
        instance = through.proxy_model(**{self.from_foreign_key: self.instance, self.to_foreign_key: value})
        instance.identifying_db_fields = [self.from_foreign_key, self.to_foreign_key]
        return instance

    def _add_object(self, child: Type["Model"]) -> None:
        self.refs.append(self.expand_relationship(child))

    async def add(self, child: "Model") -> Optional["Model"]:
        """
        Adds a child to the model as a list

        . Validates the type of the child being added to the relationship and raises error for
        if the type is wrong.
        . Checks if the middle table already contains the record being added. Raises error if yes.
        """
        if not isinstance(child, (self.to, self.to.proxy_model, self.through, self.through.proxy_model)):
            raise RelationshipIncompatible(f"The child is not from the types '{self.to.__name__}', '{self.through.__name__}'.")
        child = self.expand_relationship(child)

        try:
            async with child.database.transaction():
                return await child.save(force_save=True)
        except Exception:
            pass
        return None

    async def remove(self, child: Optional["Model"]=None) -> None:
        """Removes a child from the list of many to many.

        . Validates if there is a relationship between the entities.
        . Removes the field if there is
        """
        if self.reverse:
            fk = self.through.meta.fields_mapping[self.from_foreign_key]
        else:
            fk = self.through.meta.fields_mapping[self.to_foreign_key]
        if child is None:
            if fk.unique:
                try:
                    child = await self.get()
                except ObjectNotFound:
                    raise RelationshipNotFound(detail="no child found") from None
            else:
                raise RelationshipNotFound(detail="no child specified")
        if not isinstance(child, (self.to, self.to.proxy_model, self.through, self.through.proxy_model)):
            raise RelationshipIncompatible(f"The child is not from the types '{self.to.__name__}', '{self.through.__name__}'.")
        child = cast("Model", self.expand_relationship(child))
        count = await child.query.filter(sqlalchemy.and_(*child.identifying_clauses())).count()
        if count == 0:
            raise RelationshipNotFound(
                detail=f"There is no relationship between '{self.from_foreign_key}' and '{self.to_foreign_key}: {getattr(child,self.to_foreign_key).pk}'."
            )
        else:
            await child.delete()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"{self.through.__name__}"



class SingleRelation(ManyRelationProtocol):
    """
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)

    def __init__(
        self,
        *,
        to_foreign_key: str,
        to: Union[Type["Model"], Type["ReflectModel"]],
        embed_parent: Optional[Tuple[str, str]]=None,
        refs: Any = (),
        instance: Optional[Union["Model", "ReflectModel"]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.to = to
        self.instance = instance
        self.to_foreign_key = to_foreign_key
        self.embed_parent = embed_parent
        self.refs: Sequence[Union["Model","ReflectModel"]] = []
        if not isinstance(refs, Sequence):
            refs = [refs]
        for v in refs:
            self._add_object(v)

    def get_queryset(self) -> "QuerySet":
        # we need to check tenant every request
        queryset = self.to.meta.managers["query_related"].get_queryset()
        queryset.embed_parent = self.embed_parent
        return queryset

    def expand_relationship(self, value: Any) -> Any:
        target = self.to

        if isinstance(value, (target, target.proxy_model)):
            return value
        related_columns = self.to.meta.fields_mapping[self.to_foreign_key].related_columns.keys()
        if len(related_columns) == 1 and not isinstance(value, (dict, BaseModel)):
            value = {next(iter(related_columns)): value}
        if isinstance(value, dict):
            for key in related_columns:
                if value.get(key) is None:
                    return None
        else:
            for key in related_columns:
                if getattr(value, key, None) is None:
                    return None
        instance = target.proxy_model(**value)
        instance.identifying_db_fields = related_columns
        return instance

    def _add_object(self, child: Type["Model"]) -> None:
        self.refs.append(self.expand_relationship(child))

    def __getattr__(self, item: Any) -> Any:
        """
        Gets the attribute from the queryset and if it does not
        exist, then lookup in the model.
        """
        try:
            attr = getattr(self.get_queryset(), item)
        except AttributeError:
            attr = getattr(self.to, item)

        func = self.wrap_args(attr)
        return func

    def wrap_args(self, func: Any) -> Any:
        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            assert self.instance, "instance not initialized"
            fk = self.to.meta.fields_mapping[self.to_foreign_key]
            query = {}
            if not self.embed_parent or not isinstance(fk.owner.meta.fields_mapping[self.embed_parent[0]], RelationshipField):
                new_kwargs = kwargs
            else:
                new_kwargs = {}
                for key, val in kwargs.items():
                    if not self.embed_parent[1] or not key.startswith(self.embed_parent[1]):
                        new_kwargs[f"{self.embed_parent[0]}__{key}"] = val
                    else:
                        new_kwargs[_removeprefixes(key, self.embed_parent[1], "__")] = val

            for column_name in fk.get_column_names():
                related_name = fk.from_fk_field_name(fk.name, column_name)
                query[related_name] = getattr(self.instance, related_name)
            new_kwargs[self.to_foreign_key] = query
            return func(*args, **new_kwargs)

        return wrapped

    async def save_related(self) -> None:
        while self.refs:
            await self.add(self.refs.pop())

    async def add(self, child: "Model") -> Optional["Model"]:
        """
        Adds a child to the model as a list

        . Validates the type of the child being added to the relationship and raises error for
        if the type is wrong.
        . Checks if the middle table already contains the record being added. Raises error if yes.
        """
        if not isinstance(child, (self.to, self.to.proxy_model)):
            raise RelationshipIncompatible(f"The child is not from the type '{self.to.__name__}'.")

        await child.save(values={self.to_foreign_key: self.instance})
        return child

    async def remove(self, child: Optional["Model"]=None) -> None:
        """Removes a child from the list of one to many.

        . Validates if there is a relationship between the entities.
        . Removes the field if there is
        """
        fk = self.to.meta.fields_mapping[self.to_foreign_key]
        if child is None:
            if fk.unique:
                try:
                    child = await self.get()
                except ObjectNotFound:
                    raise RelationshipNotFound(detail="no child found") from None
            else:
                raise RelationshipNotFound(detail="no child specified")
        if not isinstance(child, self.to):
            raise RelationshipIncompatible(f"The child is not from the type '{self.to.__name__}'.")

        await child.save(values={self.to_foreign_key: None})

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"{self.to.__name__}"



"""
Client to interact with Edgy models and migrations.
"""

from typing import Any

import click

from edgy.cli.base import migrate as _migrate
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.option("-m", "--message", default=None, help="Revision message")
@click.option("--sql", is_flag=True, help=("Don't emit SQL to database - dump to standard output " "instead"))
@click.option(
    "--head",
    default="head",
    help=("Specify head revision or <branchname>@head to base new " "revision on"),
)
@click.option("--splice", is_flag=True, help=('Allow a non-head revision as the "head" to splice onto'))
@click.option("--branch-label", default=None, help=("Specify a branch label to apply to the new revision"))
@click.option("--version-path", default=None, help=("Specify specific path from config for version file"))
@click.option("--rev-id", default=None, help=("Specify a hardcoded revision id instead of generating " "one"))
@click.option("-x", "--arg", multiple=True, help="Additional arguments consumed by custom env.py scripts")
@click.command()
def makemigrations(
    env: MigrationEnv,
    directory: str,
    message: str,
    sql: bool,
    head: str,
    splice: bool,
    branch_label: str,
    version_path: str,
    rev_id: str,
    arg: Any,
) -> None:
    """Autogenerate a new revision file (Alias for
    'revision --autogenerate')"""
    _migrate(env.app, directory, message, sql, head, splice, branch_label, version_path, rev_id, arg)



from typing import TYPE_CHECKING, Type, Union

from pydantic import BaseModel

if TYPE_CHECKING:
    from edgy.core.db.models.model import Model


class ModelRef(BaseModel):
    __model__: Union[Type["Model"], str]



import click

from edgy.cli.base import revision as _revision
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.option("-m", "--message", default=None, help="Revision message")
@click.option(
    "--autogenerate",
    is_flag=True,
    help=("Populate revision script with candidate migration " "operations, based on comparison of database to model"),
)
@click.option("--sql", is_flag=True, help=("Don't emit SQL to database - dump to standard output " "instead"))
@click.option(
    "--head",
    default="head",
    help=("Specify head revision or <branchname>@head to base new " "revision on"),
)
@click.option("--splice", is_flag=True, help=('Allow a non-head revision as the "head" to splice onto'))
@click.option("--branch-label", default=None, help=("Specify a branch label to apply to the new revision"))
@click.option("--version-path", default=None, help=("Specify specific path from config for version file"))
@click.option("--rev-id", default=None, help=("Specify a hardcoded revision id instead of generating " "one"))
@click.command()
def revision(
    env: MigrationEnv,
    directory: str,
    message: str,
    autogenerate: bool,
    sql: bool,
    head: str,
    splice: bool,
    branch_label: str,
    version_path: str,
    rev_id: str,
) -> None:
    """Create a new revision file."""
    _revision(
        env.app,
        directory,
        message,
        autogenerate,
        sql,
        head,
        splice,
        branch_label,
        version_path,
        rev_id,
    )



from typing import TYPE_CHECKING, Type, cast

import sqlalchemy
from sqlalchemy.exc import DBAPIError, ProgrammingError

from edgy.core.connection.database import Database
from edgy.exceptions import SchemaError

if TYPE_CHECKING:
    from edgy import Registry


class Schema:
    """
    All the schema operations object.

    All the operations regarding a schema are placed in one object
    """

    def __init__(self, registry: Type["Registry"]) -> None:
        self.registry = registry

    def get_default_schema(self) -> str:
        """
        Returns the default schema which is usually None
        """
        return cast("str", self.registry.engine.dialect.default_schema_name)

    async def activate_schema_path(self, database: Database, schema: str, is_shared: bool = True) -> None:
        path = "SET search_path TO %s, shared;" % schema if is_shared else "SET search_path TO %s;" % schema
        expression = sqlalchemy.text(path)
        await database.execute(expression)

    async def create_schema(self, schema: str, if_not_exists: bool = False) -> None:
        """
        Creates a model schema if it does not exist.
        """

        def execute_create(connection: sqlalchemy.Connection) -> None:
            try:
                connection.execute(
                    sqlalchemy.schema.CreateSchema(name=schema, if_not_exists=if_not_exists)  # type: ignore
                )
            except ProgrammingError as e:
                raise SchemaError(detail=e.orig.args[0]) from e  # type: ignore

        async with self.registry.engine.begin() as connection:
            await connection.run_sync(execute_create)
        await self.registry.engine.dispose()

    async def drop_schema(self, schema: str, cascade: bool = False, if_exists: bool = False) -> None:
        """
        Drops an existing model schema.
        """

        def execute_drop(connection: sqlalchemy.Connection) -> None:
            try:
                connection.execute(
                    sqlalchemy.schema.DropSchema(name=schema, cascade=cascade, if_exists=if_exists)  # type: ignore
                )
            except DBAPIError as e:
                raise SchemaError(detail=e.orig.args[0]) from e  # type: ignore

        async with self.registry.engine.begin() as connection:
            await connection.run_sync(execute_drop)
        await self.registry.engine.dispose()




from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Type,
    Union,
)

from edgy.core.db.fields.base import BaseField
from edgy.core.db.fields.factories import FieldFactory

if TYPE_CHECKING:
    from edgy import Model, ReflectModel


class ConcreteExcludeField(BaseField):
    def __init__(self, **kwargs: Any):
        kwargs["exclude"] = True
        kwargs["null"] = True
        kwargs["primary_key"] = False
        return super().__init__(
            **kwargs,
        )

    def clean(self, name: str, value: Any, for_query: bool = False) -> Dict[str, Any]:
        """remove any value from input"""
        return {}

    def to_model(self, name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        """remove any value from input and raise when setting an attribute"""
        if phase == "set":
            raise AttributeError("field is excluded")
        return {}

    def __get__(self, instance: Union["Model", "ReflectModel"], owner: Any = None) -> None:
        raise AttributeError("field is excluded")


class ExcludeField(FieldFactory, Type[None]):
    """
    Meta field that masks fields
    """

    _bases = (ConcreteExcludeField,)
    _type: Any = Any



from typing import Any, Dict, Set, Type, Union

from edgy.core.db.models.base import EdgyBaseReflectModel
from edgy.core.db.models.mixins import DeclarativeMixin
from edgy.core.db.models.row import ModelRow
from edgy.exceptions import ObjectNotFound, RelationshipNotFound
from edgy.protocols.many_relationship import ManyRelationProtocol


class Model(ModelRow, DeclarativeMixin):
    """
    Representation of an Edgy `Model`.

    This also means it can generate declarative SQLAlchemy models
    from anywhere by calling the `Model.declarative()` function.

    **Example**

    ```python
    import edgy
    from edgy import Database, Registry

    database = Database("sqlite:///db.sqlite")
    models = Registry(database=database)


    class User(edgy.Model):
        '''
        The User model to be created in the database as a table
        If no name is provided the in Meta class, it will generate
        a "users" table for you.
        '''

        id: int = edgy.IntegerField(primary_key=True)
        is_active: bool = edgy.BooleanField(default=False)

        class Meta:
            registry = models
    ```
    """
    class Meta:
        abstract = True

    async def update(self, **kwargs: Any) -> Any:
        """
        Update operation of the database fields.
        """
        await self.signals.pre_update.send_async(self.__class__, instance=self)

        # empty updates shouldn't cause an error
        if kwargs:
            kwargs = self._update_auto_now_fields(kwargs, self.fields)
            expression = self.table.update().values(**kwargs).where(*self.identifying_clauses())
            await self.database.execute(expression)
        await self.signals.post_update.send_async(self.__class__, instance=self)

        # Update the model instance.
        for key, value in kwargs.items():
            setattr(self, key, value)

        for field in self.meta.fields_mapping.keys():
            _val = self.__dict__.get(field)
            if isinstance(_val, ManyRelationProtocol):
                _val.instance = self
                await _val.save_related()
        return self

    async def delete(self) -> None:
        """Delete operation from the database"""
        await self.signals.pre_delete.send_async(self.__class__, instance=self)

        expression = self.table.delete().where(*self.identifying_clauses())
        await self.database.execute(expression)

        await self.signals.post_delete.send_async(self.__class__, instance=self)

    async def load(self) -> None:
        # Build the select expression.

        expression = self.table.select().where(*self.identifying_clauses())

        # Perform the fetch.
        row = await self.database.fetch_one(expression)
        # check if is in system
        if row is None:
            raise ObjectNotFound("row does not exist anymore")
        # Update the instance.
        self.__dict__.update(self.transform_input(dict(row._mapping), phase="load"))

    async def _save(self, **kwargs: Any) -> "Model":
        """
        Performs the save instruction.
        """
        expression = self.table.insert().values(**kwargs)
        autoincrement_value = await self.database.execute(expression)
        transformed_kwargs = self.transform_input(kwargs, phase="post_insert")
        for k, v in transformed_kwargs.items():
            setattr(self, k, v)

        # sqlalchemy supports only one autoincrement column
        if autoincrement_value:
            column = self.table.autoincrement_column
            if column is not None:
                setattr(self, column.key, autoincrement_value)
        for field in self.meta.fields_mapping.keys():
            _val = self.__dict__.get(field)
            if isinstance(_val, ManyRelationProtocol):
                _val.instance = self
                await _val.save_related()
        return self

    async def save_model_references(self, model_references: Any, model_ref: Any = None) -> None:
        """
        If there is any ModelRef declared in the model, it will generate the subsquent model
        reference records for that same model created.
        """

        for reference in model_references:
            if isinstance(reference, dict):
                model: Type["Model"] = self.meta.model_references[model_ref].__model__  # type: ignore
            else:
                model: Type["Model"] = reference.__model__  # type: ignore

            if isinstance(model, str):
                model = self.meta.registry.models[model]  # type: ignore

            # If the reference did come in a dict format
            # It is necessary to convert into the original ModelRef.
            if isinstance(reference, dict):
                reference = self.meta.model_references[model_ref](**reference)  # type: ignore

            foreign_key_target_field = None
            for name, foreign_key in model.meta.foreign_key_fields.items():
                if foreign_key.target == self.__class__:
                    foreign_key_target_field = name

            if not foreign_key_target_field:
                raise RelationshipNotFound(
                    f"There was no relationship found between '{model.__class__.__name__}' and {self.__class__.__name__}"
                )

            data = reference.model_dump(exclude={"__model__"})
            data[foreign_key_target_field] = self
            await model.query.create(**data)

    def update_model_references(self, **kwargs: Any) -> Any:
        model_refs_set: Set[str] = set()
        model_references: Dict[str, Any] = {}

        for name, value in kwargs.items():
            if name in self.meta.model_references:
                model_references[name] = value
                model_refs_set.add(name)

        for value in model_refs_set:
            kwargs.pop(value)

        return kwargs, model_references

    async def save(
        self,
        force_save: bool = False,
        values: Dict[str, Any] = None,
        **kwargs: Any,
    ) -> Union["Model", Any]:
        """
        Performs a save of a given model instance.
        When creating a user it will make sure it can update existing or
        create a new one.
        """
        await self.signals.pre_save.send_async(self.__class__, instance=self)

        extracted_fields = self.extract_db_fields()

        for pkcolumn in self.__class__.pkcolumns:
            # should trigger load in case of identifying_db_fields
            if getattr(self, pkcolumn, None) is None and self.table.columns[pkcolumn].autoincrement:
                extracted_fields.pop(pkcolumn, None)
                force_save = True

        if force_save:
            if values:
                extracted_fields.update(values)
            # force save must ensure a complete mapping
            validated_values = self._extract_values_from_field(
                extracted_values=extracted_fields, is_partial=False
            )
            kwargs = self._update_auto_now_fields(values=validated_values, fields=self.fields)
            kwargs, model_references = self.update_model_references(**kwargs)
            await self._save(**kwargs)
        else:
            # Broadcast the initial update details
            # Making sure it only updates the fields that should be updated
            # and excludes the fields aith `auto_now` as true
            validated_values = self._extract_values_from_field(
                extracted_values=extracted_fields if values is None else values,
                is_update=True,
                is_partial=values is not None,
            )
            kwargs, model_references = self.update_model_references(**validated_values)
            update_model = {k: v for k, v in validated_values.items() if k in kwargs}

            await self.signals.pre_update.send_async(self.__class__, instance=self, kwargs=update_model)
            await self.update(**update_model)

            # Broadcast the update complete
            await self.signals.post_update.send_async(self.__class__, instance=self)

        # Save the model references
        if model_references:
            for model_ref, references in model_references.items():
                await self.save_model_references(references or [], model_ref=model_ref)

        # Refresh the results
        if any(field.server_default is not None for name, field in self.fields.items() if name not in extracted_fields):
            await self.load()

        await self.signals.post_save.send_async(self.__class__, instance=self)
        return self


class ReflectModel(Model, EdgyBaseReflectModel):
    """
    Reflect on async engines is not yet supported, therefore, we need to make a sync_engine
    call.
    """

    class Meta:
        abstract = True



from typing import TYPE_CHECKING, List, Union

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from edgy import Model


class ConfigMarshall(TypedDict, total=False):
    """A TypedDict for configuring Marshall behaviour."""

    model: Union["Model", str]
    """The model from there the marshall will read from."""

    fields: Union[List[str], None]
    """A list of fields to be serialized"""

    exclude: Union[List[str], None]
    """A list of fields to be excluded from the serialization."""



import click

from edgy.cli.base import init as _init
from edgy.cli.env import MigrationEnv


@click.option(
    "-d",
    "--directory",
    default=None,
    help=('Migration script directory (default is "migrations")'),
)
@click.option("-t", "--template", default=None, help=('Repository template to use (default is "flask")'))
@click.option(
    "--package",
    is_flag=True,
    help=("Write empty __init__.py files to the environment and " "version locations"),
)
@click.command(name="init")
def init(env: MigrationEnv, directory: str, template: str, package: bool) -> None:
    """Creates a new migration repository."""
    _init(env.app, directory, template, package)



from typing import Union

import click

from edgy.utils.inspect import InspectDB


@click.option(
    "--database",
    required=True,
    help=("Connection string. Example: postgres+asyncpg://user:password@localhost:5432/my_db"),
)
@click.option(
    "--schema",
    default=None,
    help=("Database schema to be applied."),
)
@click.command()
def inspect_db(
    database: str,
    schema: Union[str, None] = None,
) -> None:
    """
    Inspects an existing database and generates the Edgy reflect models.
    """
    inspect_db = InspectDB(database=database, schema=schema)
    inspect_db.inspect()



import functools
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Dict, Tuple, Type, Union, cast

from edgy.core.db.fields.base import RelationshipField
from edgy.core.db.fields.foreign_keys import BaseForeignKeyField
from edgy.protocols.many_relationship import ManyRelationProtocol

if TYPE_CHECKING:
    from edgy import Model, ReflectModel

class RelatedField(RelationshipField):
    """
    When a `related_name` is generated, creates a RelatedField from the table pointed
    from the ForeignKey declaration and the the table declaring it.
    """

    def __init__(
        self,
        *,
        foreign_key_name: str,
        related_from: Union[Type["Model"], Type["ReflectModel"]],
        **kwargs: Any
    ) -> None:
        self.foreign_key_name = foreign_key_name
        self.related_from = related_from
        super().__init__(
            inherit=False,
            exclude=True,
            deprecated=False,
            __type__=Any,
            annotation=Any,
            column_type=None,
            null=True,
            **kwargs
        )

    @property
    def related_to(self) ->  Union[Type["Model"], Type["ReflectModel"]]:
        return self.owner

    @property
    def related_name(self) ->  str:
        return self.name

    def to_model(self, field_name: str, value: Any, phase: str = "") -> Dict[str, Any]:
        """
        Meta field
        """
        if isinstance(value, ManyRelationProtocol):
            return {field_name: value}
        return {field_name: self.get_relation(refs=value)}

    def __get__(self, instance: "Model", owner: Any = None) -> ManyRelationProtocol:
        if instance:
            if self.name not in instance.__dict__:
                instance.__dict__[self.name] = self.get_relation()
            if instance.__dict__[self.name].instance is None:
                instance.__dict__[self.name].instance = instance
            return instance.__dict__[self.name]  # type: ignore
        raise ValueError("missing instance")

    def __set__(self, instance: "Model", value: Any) -> None:
        relation = self.__get__(instance)
        if not isinstance(value, Sequence):
            value = [value]
        for v in value:
            relation._add_object(v)

    @functools.cached_property
    def foreign_key(self) -> BaseForeignKeyField:
        return cast(BaseForeignKeyField, self.related_from.meta.fields_mapping[self.foreign_key_name])

    def traverse_field(self, path: str) -> Tuple[Any, str, str]:
        return self.foreign_key.reverse_traverse_field(path)

    def get_relation(self, **kwargs: Any) -> ManyRelationProtocol:
        return self.foreign_key.get_relation(**kwargs)

    def is_cross_db(self) -> bool:
        return self.foreign_key.is_cross_db()

    def clean(self, name: str, value: Any, for_query: bool = False) -> Dict[str, Any]:
        if not for_query:
            return {}
        return self.related_to.meta.pk.clean("pk", value, for_query=for_query)  # type: ignore

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return f"({self.related_to.__name__}={self.related_name})"
